local modName = "Totem Ranges"
local parent = HudMap
local L = LibStub("AceLocale-3.0"):GetLocale("HudMap")
local mod = HudMap:NewModule(modName, "AceEvent-3.0")

--[[ Default upvals 
     This has a slight performance benefit, but upvalling these also makes it easier to spot leaked globals. ]]--
local _G = _G.getfenv(0)
local wipe, type, pairs, tinsert, tremove, tonumber = _G.wipe, _G.type, _G.pairs, _G.tinsert, _G.tremove, _G.tonumber
local math, math_abs, math_pow, math_sqrt, math_sin, math_cos, math_atan2 = _G.math, _G.math.abs, _G.math.pow, _G.math.sqrt, _G.math.sin, _G.math.cos, _G.math.atan2
local error, rawset, rawget, print = _G.error, _G.rawset, _G.rawget, _G.print
local tonumber, tostring = _G.tonumber, _G.tostring
local getmetatable, setmetatable, pairs, ipairs, select, unpack = _G.getmetatable, _G.setmetatable, _G.pairs, _G.ipairs, _G.select, _G.unpack
--[[ -------------- ]]--

local db

local totemRanges = {
	-- Fire
	{
		[3599]  = 30,	-- Searing
		[8181]  = 30,	-- Frost Resist
		[8190]  = 8,	-- Magma
		[8227]  = 30,	-- Flametongue
		[30706] = 40,	-- Totem of Wrath
	},
	
	-- Earth
	{
		[2484]  = 10,	-- Earthbind
		[5730]  = 8,	-- Stoneclaw
		[8071]  = 30,	-- Stoneskin
		[8075]  = 30,	-- Strength of Earth
		[8143]  = 30,	-- Tremor
	},	
	
	-- Water
	{
		[5675]  = 30,	-- Mana Spring
		[8170]  = 30,	-- Cleansing
		[8184]  = 30,	-- Fire Resist
		[16190] = 30,	-- Mana Tide
		[5394]  = 30,	-- Healing Stream
	},
	
	-- Wind
	{
		[3738] = 30,	-- Wrath of Air
		[8177]  = 20,	-- Grounding
		[8512]  = 30,	-- Windfury
		[10595] = 30,	-- Nature Resist
	},
	
	--[[ Heart
	{
		[70575] = 9001
	}
	]]--
}

local options = {
	type = "group",
	name = L["Totems"],
	args = {
		enable = {
			type = "toggle",
			name = L["Enable Totem Ranges"],
			order = 100,
			get = function()
				return db.enable
			end,
			set = function(info, v)
				db.enable = v
			end
		},
		myTotems = {
			type = "group",
			name = L["My Totems"],
			args = {
				enable = {
					type = "toggle",
					name = L["Enable My Totems"],
					order = 100,
					get = function()
						return db.myTotems.enable
					end,
					set = function(info, v)
						db.myTotems.enable = v
					end
				},				
			}
		},
		partyTotems = {
			type = "group",
			name = L["Party Totems"],
			args = {
				enable = {
					type = "toggle",
					name = L["Enable Party Totems"],
					order = 101,
					get = function()
						return db.partyTotems.enable
					end,
					set = function(info, v)
						db.partyTotems.enable = v
					end
				}
			}
		}
	}
}
local totemNames = {}
local totemGUIDs = {}
local totemCircles = {}
local groups = {
	L["Fire Totems"],
	L["Earth Totems"],
	L["Water Totems"],
	L["Air Totems"]
}

local defaults = {
	profile = {
		enable = true,
		myTotems = { enable = true },
		partyTotems = { enable = false }
	}
}

local function getSlotForTotem(totem)
	for index, set in ipairs(totemRanges) do
		for id, range in pairs(set) do
			if id == totem then
				return index
			end
		end
	end
end

for _, key in ipairs({"myTotems", "partyTotems"}) do
	for index, v in ipairs(totemRanges) do
		local optGroup = {
			name = groups[index],
			type = 'group',
			args = {},
			disabled = function()
				return not db[key].enable
			end
		}
		for id, range in pairs(v) do
			local n = GetSpellInfo(id)
			local sid = key .. n
			defaults.profile[key][n] = true
			optGroup.args[n:gsub(" ", "")] = {
				name = n,
				type = "toggle",
				get = function()
					return db[key][n]
				end,
				set = function(info, v)
					db[key][n] = v
				end
			}		
		end
		options.args[key].args[groups[index]:gsub(" ", "")] = optGroup
	end
end

local colors = {
	{1, 36/255, 0, 0.7},
	{200/255, 121/255, 43/255, 0.7},
	{0, 138/255, 1, 0.7},
	{150/255, 1, 0, 0.7}	
}

local effectiveTotemRanges = {}

local talents = {
	earths_grasp = {totem_ids = {[2484] = 1}, range_per_point = 10, tab_index = 2, talent_index = 2}
}
local offsets = {
	{-1, 1},
	{1, 1},
	{1, -1},
	{-1, -1},
}

function mod:OnInitialize()
	self.db = parent.db:RegisterNamespace(modName, defaults)
	db = self.db.profile
	parent:RegisterModuleOptions(modName, options, modName)
end

function mod:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("PLAYER_TOTEM_UPDATE")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateTotemInformation")
	self:UpdateTotemInformation()
end

local function findRangeAndSlot(targetID)
	local range, slot
	for s, totems in ipairs(totemRanges) do
		for id, r in pairs(totems) do
			if targetID == id then
				slot = s
				range = r
				break
			end
		end
		if range then break end
	end
	return range, slot
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, ...)
	if event == "SPELL_CAST_SUCCESS" then
		local totemName = GetSpellInfo(spellID)
		if totemNames[totemName] and (UnitInRaid(sourceName) or UnitInParty(sourceName)) then
			local targetID = totemNames[totemName]
			local range, slot = findRangeAndSlot(targetID)
			if not range or not slot then return end
			local x, y = HudMap:GetUnitPosition(sourceName)
			if UnitIsUnit(sourceName, "player") then
				if not db.myTotems.enable then return end
				if not db.myTotems[totemName] then return end
				local offset = offsets[slot]		
				local bearing = -GetPlayerFacing()
				local angle = math_atan2(offset[1], offset[2])
				local hyp = 2.9 -- math.sqrt(8)
				local nx, ny = math_sin(angle + bearing), math_cos(angle + bearing)
				x = x + (nx * hyp)
				y = y + (ny * hyp)			
				range = effectiveTotemRanges[totemName]
			else
				if not db.partyTotems.enable then return end
				if not db.partyTotems[totemName] then return end
			end
			local r, g, b, a = unpack(colors[slot])
			local tex = range > 25 and "radius" or "radius_lg"
			local key = sourceGUID .. ":" .. slot
			for k, v in pairs(totemGUIDs) do
				if v == key then
					totemGUIDs[k] = nil
				end
			end
			if totemCircles[key] then totemCircles[key]:Free() end
			totemCircles[key] = HudMap:PlaceRangeMarker(tex, x, y, range, 300, r, g, b, a):Rotate(360 * (slot % 2 == 0 and 1 or -1), 175+(slot*3)):Appear()
		end
	elseif event == "SPELL_SUMMON" then
		local totemName = GetSpellInfo(spellID)
		if totemNames[totemName] and (UnitInRaid(sourceName) or UnitInParty(sourceName)) then
			local totemName = GetSpellInfo(spellID)
			local targetID = totemNames[totemName]
			local range, slot = findRangeAndSlot(targetID)
			if not range or not slot then return end
			totemGUIDs[destGUID] = sourceGUID .. ":" .. slot
		end
	elseif event == "UNIT_DIED" then
		if totemGUIDs[destGUID] then
			local dot = totemCircles[totemGUIDs[destGUID]]
			if dot then
				totemCircles[totemGUIDs[destGUID]] = dot:Free()
				totemGUIDs[destGUID] = nil
			end
		end
	end
end

function mod:UpdateTotemInformation()
	wipe(effectiveTotemRanges)
	wipe(totemNames)
	for set, totems in ipairs(totemRanges) do
		for id, range in pairs(totems) do 
			local name, _, icon = GetSpellInfo(id)
			totemNames[name] = id
			for talent_name, talent_data in pairs(talents) do
				if talent_data.totem_ids[id] then
					local points = select(5, GetTalentInfo(talent_data.tab_index, talent_data.talent_index))
					range = range + (points * talent_data.range_per_point)
				end
			end
			effectiveTotemRanges[name] = range
		end
	end
end

function mod:PLAYER_TOTEM_UPDATE(event, slot)
	local haveTotem, totemName, startTime, duration, icon = GetTotemInfo(slot)
	if startTime == 0 then
		local key = UnitGUID("player") .. ":" .. slot
		if totemCircles[key] then
			totemCircles[key] = totemCircles[key]:Free()
		end
	end
end
